function [X_min, f_min, bFoundMin] = evolution(fitnessFct, nDim, SeekArea, nParents, nOffspring, bRecombine, bParentsDying, sigmaInit, sigma_min, sigma_max, tau_0, tau_1, eps_min, eps_min_q, q, nMaxGenerations, bDisplay)
%function [X_min, f_min, bFoundMin] = evolution(fitnessFct,
%												nDim, SeekArea,
%												nParents, nOffspring,
%												bRecombine, bParentsDying,
%												sigmaInit, sigma_min, sigma_max, 
%												tau_0, tau_1,
%												eps_min, eps_min_q, q,
%												nMaxGenerations,
%												bDisplay)
%
% Versucht minimalen Wert von fitnessFct mithilfe der im Skript
% "Fuzzy Logik, Neuronale Netze und Evolutionre Algorithmen", J. Adamy,
% Abschnitt 4.3 vorgestellten Evolutionsstrategie zu ermitteln.
%
% Eric Lenz, 10.01.'06
%
% Auskommentierte Version 18.01.'06
%
%
% Parameter:
%	fitnessFct : 	Inlinefunktion, die einen Zeilenvektor mit
%					nDim Elementen als Eingangswert erwartet
%					und einen skalaren Wert zurckgibt.
%
%	nDim : 			Anzahl der Variablen der Fitness-Funktion (Gtewertfunktion)
%	SeekArea : 		(nDim x 2)-Matrix, die in jeder Dimension eine linke und
%					rechte Grenze fr den Suchraum festlegt
%
%	nParents :		Anzahl der Eltern
%	nOffspring : 	Anzahl der Nachkommen
%	bRecombine :	1, wenn Rekombination (X: diskret, sigma: Mittelwert)
%					0, wenn Eltern geklont werden
%	bParentsDying : 1, wenn Eltern aussterben
%					0, wenn Eltern nicht aussterben
%
%	sigmaInit : 	Sigma-Wert der ersten Generation (Ist fr alle Individuen
%					und alle Dimensionen gleich)
%
%	sigma_min :		Minimaler Wert, den ein sigma annehmen kann
%					(= '-Inf', wenn keine Beschrnkung erfolgen soll)
%	sigma_max :		Maximaler Wert, den ein sigma annehmen kann
%					(= 'Inf', wenn keine Beschrnkung erfolgen soll)
%	tau_0 : 		Nachfolgerbeiwert (Skript S. 303)
%	tau_1 : 		Dimensionsbeiwert (Skript S. 303)
%
%	eps_min : 		Abbruchkriterium bei Betrachtung einer Generation
%					(= 'Inf', wenn dieses Kriterium nicht benutzt werden soll)
%	eps_min_q :		Abbruchkriterium bei Betrachtung ber mehrere
%					Generationen
%					(= 'Inf', wenn dieses Kriterium nicht benutzt werden soll)
%	q :				Anzahl der Generationen, die in das Abbruchkriterien
%					einbezogen werden
%	nMaxGenerations:Maximale Anzahl an Generationen
%
%	bDisplay :		1, wenn ausfhrlichere Bildschirmausgabe erfolgen soll
%
%
% Rckgabewerte:
%	X_min :			Koordinaten des gefundenen "Minimums"
%	f_min :			"minimaler" Funktionswert
%	bFoundMin :		=1, wenn die Abbruchbedingungen eps_min und eps_min_q
%					    erfllt wurden
%					=0, wenn maximale Anzahl der Generationen erreicht
%					    wurde


	% Allgemein
	% Ein Individuum ist in zwei Teile geteilt: X-Werte und Sigma-Werte
	% Dies sind jeweils Zeilenvektoren und werden mit mehreren Individuen zu Matrizen
	% zusammengefasst. Es gibt also immer ein Paar von Matizen, z.B.
	% Offspring_X und Offspring_S, die die Parameter aller Nachkommen einer
	% Generation (spaltenweise) beinhalten.

	
	% berprfung Eingaben

	% q muss mindestens 1 sein. Dieses Abbruchkriterium wird durch Setzen
	% von eps_min_q auf "inf" auer Kraft gesetzt
	if (q == 0)
		q = 1;
		eps_min_q = inf;
	end
	
	if ( (nOffspring < nParents) && (bParentsDying ~= 0) )
		error('Wenn Eltern sterben, darf es nicht weniger Nachkomme als Eltern geben!')	
	end
	
	
	
	% Initialisieren von X_min, und f_min
	X_min = 0;
	f_min = 0;

	% Vektor, der die besten Werte aus q Generationen speichert.
	% Mit inf vorbesetzen (siehe auch Funktion selection)
	F_best = ones([q 1]) * Inf;

	% Erzeugen der ersten Generation
	% Die Individuen werden dabei gleichverteilt zufllig ber den Suchraum
	% verteilt. Die Sigma-Werte bekommen alle den selben Startwert
	% sigmaInit
	[Parents_X Parents_S] = createFirstGeneration(nParents, nDim, SeekArea, sigmaInit);

	% Schleife ber maximale Anzahl von Generationen
	% (Sollte eine der epsilon-Abbruchbedingungen greifen wird diese ber
	% eine break-Anweisung verlassen.)
	for gen = 1 : nMaxGenerations
		if (bDisplay ~= 0)
			disp(' ')
			disp(' ')
			disp(' ')
		end
		
		% Generationsnummer wird auf jeden Fall ausgegeben
		disp(['Generation : ' num2str(gen)])
		
		if (bDisplay ~= 0)
			disp('------------')
			disp(' ')
		end
	
		% Je nach Auswahl Rekombination oder Klonen der nchsten Nachkommen
		% durchfhren
		if (bRecombine ~= 0)
			[Offspring_X Offspring_S] = recombine(nDim, nOffspring, nParents, Parents_X, Parents_S);
		else
			[Offspring_X Offspring_S] = clone(nDim, nOffspring, nParents, Parents_X, Parents_S);		
		end
	
		% Nachkommen mutieren
		Offspring_S = mutateS(nDim, nOffspring, Offspring_S, tau_0, tau_1, sigma_min, sigma_max);
		Offspring_X = mutateX(nDim, nOffspring, Offspring_X, Offspring_S, SeekArea);

		% Falls Eltern sterben, gehren nur noch die Nachkommen zur
		% Bevlkerung die dann selektiert wird, ansonsten besteht die
		% Bevlkerung aus Eltern und Nachkommen
		if (bParentsDying ~= 0)
			People_X = [Offspring_X];
			People_S = [Offspring_S];
			nPeople = nOffspring;
		else
			People_X = [Parents_X Offspring_X];
			People_S = [Parents_S Offspring_S];
			nPeople = nParents + nOffspring;
		end

		if (bDisplay ~= 0)
			disp('Eltern:')
			disp('X:')
			disp(num2str(Parents_X))
			disp('Sigma:')
			disp(num2str(Parents_S))
			disp(' ')
			disp('Nachkommen:')
			disp('X:')
			disp(num2str(Offspring_X))
			disp('Sigma:')
			disp(num2str(Offspring_S))
			disp(' ')	
			disp(' ')	
		end	%(bDisplay ~= 0)
	
		% Beste Individuen ermitteln und Abbruchbedingung prfen
		[Parents_X Parents_S bFoundMin F_best] = selection(fitnessFct, nDim, nParents, People_X, People_S, nPeople, eps_min, eps_min_q, F_best, q, bDisplay);
	
		% Falls Abbruchbedingung erfllt Schleife abbrechen
		if (bFoundMin == 1)
			break;
		end
	end	%gen = 1 : nMaxGenerations

	X_min = Parents_X(:, 1);
	f_min = fitnessFct(X_min);

% end function evolution
% -------------------------------------------------------------------------



% function clone
%
% Whlt fr jeden der nOffspring Klone zufllig ein Elter aus und gibt dem
% Klon dieselben Parameter wie dem Elter
function [Offspring_X, Offspring_S] = clone(nDim, nOffspring, nParents, Parents_X, Parents_S)

	% Erzeugen von Zeilenvektor mit Zahlen zwischen 0 und nParents
	% Dazu wir Zufallszahl von ]0 1[ mit nParents multipliziert und dann
	% auf den nchsthheren Integer gerundet
	Parent = ceil( rand([nOffspring, 1]) * nParents )

	% Kopieren der Elter-Parameter fr jeden Klon
	for i = 1 : nOffspring
		Offspring_X(:, i) = Parents_X(:, Parent(i));
		Offspring_S(:, i) = Parents_S(:, Parent(i));
	end
	
% end function clone
% -------------------------------------------------------------------------



% function createFirstGeneration
%
% Erzeugt nParents Individuen die sich gleichverteilt zufllig in dem
% Suchraum befinden. Alle Sigma-Werte werden mit sigmaInit vorbelegt
function [Parents_X, Parents_S] = createFirstGeneration(nParents, nDim, SeekArea, sigmaInit)

	% Alle Sigma-Werte gleich sigmaInit
	Parents_S = ones([nDim, nParents]) * sigmaInit;
	
	Parents_X = rand([nDim, nParents]);
	
	for k = 1 : nDim
		Parents_X(k, :) = Parents_X(k, :) * ( SeekArea(k, 2) - SeekArea(k, 1) );
		Parents_X(k, :) = Parents_X(k, :) + SeekArea(k, 1);
	end
	
% end function createFirst Generation
% -------------------------------------------------------------------------
	

	
% function getCouples
%
% Bildet aus nParents Eltern nOffspring Paare
% Gibt eine (nOffspring x 2)-Matrix zurck, deren Zeilen die Nummern der
% Eltern je eines Nachkommens enthlt
function Couples = getCouples(nParents, nOffspring)

	% Erzeugt eine (nOffspring x 2)-Matrix mit gleichverteilten
	% Zufallszahlen im Intervall ]0 1[
	Couples = rand([nOffspring 2]);
	% Fr den ersten Partner gibt es nParents Mglichkeiten, fr den
	% zweiten dann nur noch (nParents - 1)
	Couples(:, 1)  = Couples(:, 1) * (nParents);
	Couples(:, 2)  = Couples(:, 2) * (nParents - 1);
	
	% Runden auf den nchsthchsten Integerwert
	Couples = ceil( Couples );

	% Wenn der zweite Partner dieselbe oder eine grere Nummer als der
	% erste besitzt, wird 1 dazuaddiert. So wird eine Gleichverteilung
	% der Wahrscheinlichkeit ber die verbleibenden Partner erreicht.
	for r = 1 : nOffspring
		if (Couples(r, 2) >= Couples(r, 1))
			Couples(r, 2) = Couples(r, 2) + 1;
		end
	end
	
% end function getCouples
% -------------------------------------------------------------------------



% funtion Dominant
%
% Gibt eine (nDim x 2)-Matrix zurck, die fr jede Dimension der X-Werte
% festlegt, ob der erste oder zweite Partner dominiert. (Fr diskrete
% Rekombination.)
% Dominiert der erste Partener lautet die entsprechende Zeile [1 0],
% dominiert der zweite [0 1]
function Dominant = getDominant(nDim)

	% Erzeugt erste Spalte mit gleichverteilten Einsen und Nullen durch
	% erzeugen einer geleichverteilten Zufallszahl ]0 1[ und kaufmnnischem
	% Runden
	Dominant = round( rand([nDim 1]) );
	
	% Zweite Spalte wird mit dem Gegenteil der ersten gefllt
	Dominant(:, 2) = 1 - Dominant(:, 1);

% end function get Dominiant
% -------------------------------------------------------------------------



% function mutateS
%
% Fhrt eine Mutation der Sigma-Werte der Nachkommen durch.
% Die S-Werte beleiben dabei auf den Bereich [sigma_min sigma_max]
% beschrnkt.
% Die Mutation erfolgt nach der im Skript angegeben Formel.
function Offspring_S_mutate = mutateS(nDim, nOffspring, Offspring_S, tau_0, tau_1, sigma_min, sigma_max)

	% Die beiden Teilfaktoren werden laut Skript bestimmt
	OffspringFactor = exp( tau_0 * randn([nOffspring, 1]) );
	DimFactor = exp( tau_1 * randn([nDim, 1]) );

	% Die Faktor-Matrix ensteht durch die Vektormultiplikation von
	% DimFactor und der transponierten von OffspringFactor und beinhaltet
	% so fr jeden Sigma-Wert jedes Individuums einen Zufallsfaktor
	Factors = DimFactor * OffspringFactor';

	% Mutation wird durch elementweises Multiplitzieren jedes Sigma-Wertes
	% mit den eben bestimmten Zufallsfaktoren druchgefhrt.
	Offspring_S_mutate = Offspring_S .* Factors;

	% berprfen, ob alle Werte im vorgegebenen Bereich liegen, falls nicht wird der
	% entsprechende Wert auf den Rand des Bereiches zurckgesetzt.
	for i = 1 : nOffspring
		for k = 1 : nDim
			if ( Offspring_S_mutate(k, i) < sigma_min )
				Offspring_S_mutate(k, i) = sigma_min;
			elseif ( Offspring_S_mutate(k, i) > sigma_max )
				Offspring_S_mutate(k, i) = sigma_max;			
			end
		end %k = 1 : nDim
	end %i = 1 : nOffspring

% end function mutateS
% -------------------------------------------------------------------------



% function mutateX
%
% Fhrt eine Mutation der X-Werte der Nachkommen durch.
% Die X-Werte beleiben dabei auf den Suchbereich beschrnkt.
% Die Mutation erfolgt nach der im Skript angegeben Formel.
function Offspring_X_mutate = mutateX(nDim, nOffspring, Offspring_X, Offspring_S, SeekArea)

	% Matrix, die einen standard-normalverteilten Zufallsfaktor fr jede
	% Dimension jedes Individuums enthlt
	Factors = randn([nDim nOffspring]);
    
	% nderung der X-Werte durch Elementweises Multiplizieren von den
	% gerade bestimmten Zufallsfaktoren mit den Sigma-Werten
	Offspring_X_mutate = Offspring_X + Offspring_S .* Factors;

	% berprfen, ob alle Werte im Suchbereich liegen, falls nicht wird der
	% entsprechende Wert auf den Rand des Suchbereiches zurckgesetzt.
	for i = 1 : nOffspring
		for k = 1 : nDim
			if ( Offspring_X_mutate(k, i) < SeekArea(k, 1) )
				Offspring_X_mutate(k, i) = SeekArea(k, 1);
			elseif ( Offspring_X_mutate(k, i) > SeekArea(k, 2) )
				Offspring_X_mutate(k, i) = SeekArea(k, 2);
			end
		end %k = 1 : nDim
	end %i = 1 : nOffspring
	
% end function mutateX
% -------------------------------------------------------------------------



% function recombine
%
% Bildet nOffsrping Paare aus nParents Individuen und kombiniert deren
% X-Werte nach der diskreten und deren Sigma-Werte nach der intermediren
% Methode
function [Offspring_X, Offspring_S] = recombine(nDim, nOffspring, nParents, Parents_X, Parents_S)

	% Bilden der Paare
	Couples = getCouples(nParents, nOffspring);

	% Fr jeden Nachkomme werden die X- und Sigma-Werte bestimmt
	for i = 1 : nOffspring
		% Bestimmt die Dominaten Parameter fr die diskrete Kombination der
		% X-Werte
		Dominant = getDominant(nDim);
		% Durch elementweise Multiplikation mit Dominant(:, 1) werden alle
		% nichtdominanten Eintrge des ersten Partners ausgeblendet (zu Null),
		% entsprechend mit Dominant(:, 2) die des zweiten
		Offspring_X(:, i) = Parents_X( :, Couples(i, 1) ) .* Dominant(:, 1) + Parents_X( :, Couples(i, 2) ) .* Dominant(:, 2);

		% Fr Sigma-Werte einfach der Mittelwert gebildet
		Offspring_S(:, i) = 0.5 * ( Parents_S(:, Couples(i, 1)) + Parents_S(:, Couples(i, 2)) );
	end %i = 1 : nOffspring
	
% end function recombine
% -------------------------------------------------------------------------
	


% function selection
%
% Wertet die Fitness-Funktion fr alle Individuen der aktuellen Bevlkerung
% aus und gibt die Parameter der besten nParents Individuen aus. Diese bilden
% dann die Eltern der nchsten Generation.
%
% Werden die Abbruchbedingungen erfllt wird bFoundMin = 1 zurckgegeben,
% ansonsten ist bFoundMin = 0.
% F_best_new beinhaltet die besten Fitness-Werte der q zurckliegenden
% Generationen und muss bei jedem Aufruf wieder neu bergeben werden!
function [Survivors_X, Survivors_S, bFoundMin, F_best_new] = selection( fitnessFct, nDim, nParents, People_X, People_S, nPeople, eps_min, eps_min_q, F_best, q, bDisplay)

	bFoundMin = 0;

	% Die Fitness wird fr jedes Individuum bestimmt
	% Die Matrix Fitness enthlt in der ersten Spalte die Nummer der
	% Individuen, in der zweiten deren Fitness-Werte
	for i = 1 : nPeople
		Fitness(i, 1) = i;
		Fitness(i, 2) = fitnessFct( People_X(:, i) );
	end

	% Order ist die sortierte Version der Fitness-Matrix, dabei wird nach
	% der zweiten Spalte (also den Fitness-Werten) aufsteigend sortiert.
	Order = sortrows(Fitness, 2);
	% Die erste Spalte von Order enthlt nun Rangfolge der Individuen
	% anhand der Fitness-Werte. (Das beste, also das Individuum mit dem
	% kleinsten Fitness-Wert ist dabei an erster Stelle.) Die Fitnesswerte
	% werden in Order verworfen.
	Order = Order(:, 1);

	% Der beste Fitnesswert dieser Generation wird an das Ende von F_best
	% geschrieben. Alle anderen rcken um eine Stelle vor, der lteste Wert
	% fllt weg
	F_best = [F_best(2:q); Fitness(Order(1), 2)];
	%disp(num2str(Fitness(Order(1), 2)))
	% Der Vektor F_best_new wird von der Funktion wieder zurckgegeben und
	% beim nchsten Aufruf wieder bergeben. (Eine Alternative wre eine
	% globale Variable zu benutzen.)
	F_best_new = F_best;

	if (bDisplay ~= 0)
		disp(['Fitness:  ' num2str( Fitness(:, 2)' ) ])
		disp(['F_best:   ' num2str( F_best' ) ])
	end

	% Wenn der oberste Eintrag in F_best ungleich inf ist, allso mindestens q
	% Generationen ausgewertet sind, werden die Abbruchkriterien berprft
	if (F_best(1) ~= Inf)
		% Berechnung des maximalen relativen Fehlers basierend auf den
		% besten Individuen der letzten q Generationen
		eps_q = norm(    ( max(F_best) - min(F_best) ) / min(F_best)    );

		% Berechnung des maximalen relativen Fehlers basierend auf den
		% Individuen dieser Generationen
		f_best = Fitness(Order(1), 2);
		f_worst = Fitness(Order(nPeople), 2);
		eps = norm( (f_worst - f_best) / f_best );

		% Bildschirmausgabe beider realtiver Fehler
		disp(['eps_q:  ' num2str(eps_q) '        eps:  ' num2str(eps)])
	    % berprfung der Abbruchkriterien
		% Wenn Bedingungen erfllt wird bFoundMin auf 1 gesetzt
		if ( (eps_q < eps_min_q) && (eps < eps_min) )
			bFoundMin = 1;
		end
	end	%(F_best(1) ~= Inf)

	
	% Die nParents besten Indiviuen bilden die nchste Generation un werden
	% von der Funktion zurckgegeben
	for i = 1 : nParents
		Survivors_X(:, i) = People_X(:, Order(i));
		Survivors_S(:, i) = People_S(:, Order(i));	
	end
	
% end function selection
